****************************************************************************
****************************************************************************
** Name: SOLITAIRE (HP49G only)
**
** Author: John Olwoch
**
** Abstract:    The objective of the game is to move one ball over another
**              into a vacant cup, finishing off with just one ball in the
**              centre cup. The game uses ideas by Doug Cannon (the variable
**              storage system), Jim Donnelly (the key handler) and Julien
**              Meyer (sprite display).
**
**              Keys            Function
**              ----            --------
**              UpArrow, 8      Move cursor/ball up
**              DownArrow, 2    Move cursor/ball down
**              LeftArrow, 4    Move cursor/ball left
**              RightArrow, 6   Move cursor/ball right
**              5               Prepare to move a ball
**              SPC             Remove the first ball
**              ENTER           Save then Quit
**              ON              Quit
**
** Entry:	None
**
** Exit:	None
**
**     Date     Prog             Modification
**   --------   ----   -----------------------------------------------------
**   16/09/99   JO     Implemented.
**==========================================================================

****************************************************************************
*
*  Storage variables
*
*  00: Pointer to ABUFF
*  01: X position of ball
*  02: Y position of ball
*  03: Pointer to game board
*  04: Pointer to X coordinates
*  05: Pointer to Y coordinates
*
****************************************************************************
*
*  Scratch registers
*
*  R0: Pointer to variable store
*
****************************************************************************
*
*  Flags
*
*   1: Blink
*   2: Ball/space in current location
*   3: Cursor/ball movement
*   4:
*   5: Remove first ball then reset permanently
*   6: Direction 0=left or up, 1=right or down
*
****************************************************************************
****************************************************************************

ASSEMBLE
	NIBASC /HPHP49-C/
RPL

::
  AtUserStack
  RECLAIMDISP                   ( *clear text GROB* )
  TURNMENUOFF                   ( *remove menu* )

* Draw the title screen.

ASSEMBLE
	CON(5)  =DOGROB
	REL(5)  title_end
	CON(5)  32
	CON(5)  88
	NIBHEX  000000087000008700000000000004F000004F000000
	NIBHEX  00000002F100002F1000000000000AF10000AF100000
	NIBHEX  0000000EF10000EF1000000000000EF10000EF100000
	NIBHEX  FFFFFFFCFCFFFFCFCFFFF0FFFFFFF97EFFFF97EFFFF0
	NIBHEX  00000000000000000000000000000000000000000000
	NIBHEX  C30F030CFCF3C3CFCF0FF0E78F130CFCF3E7CFCF1FF0
	NIBHEX  7EC9330030607E03C833003CC0330030603C03C03300
	NIBHEX  30C0330030603C03C0330030C0330030603C03C03300
	NIBHEX  30C0330030603C03C0330070C0330030603C03C83300
	NIBHEX  E3C033003060FF03CF1F30C7C033003060FF03CF0F30
	NIBHEX  0EC0330030603C03C303000CC0330030603C03C70300
	NIBHEX  0CC0330030603C03CE03000CC0330030603C03CC1300
	NIBHEX  3CC0330030603C03C833007EC9330030603C03C03300
	NIBHEX  E78F1FFCF0603CCFC03FF0C30F0FFCF0603CCFC03FF0
	NIBHEX  00000000000000000000000000000000000000000000
	NIBHEX  FFFFFFFFFFFFFFFFFFFFF0FFFFFFFFFFFFFFFFFFFFF0
title_end
RPL
  ABUFF
  TWENTYTHREE TWELVE GROB!      ( *Display title screen* )

  ' ID SOLpar @                 ( *Recall old game if available* )
  ITE
  ZERO                          ( *This identifies a saved game* )
  ::
ASSEMBLE                        ( *Else, create a new one* )
	CON(5)	=DOEXT0
	REL(5)	BOARD_END
	NIBHEX	0022200
	NIBHEX	0022200
	NIBHEX	2222222
	NIBHEX	2222222
	NIBHEX	2222222
	NIBHEX	0022200
	NIBHEX	0022200
BOARD_END
RPL
    TOTEMPOB
    ONE                         ( *New game* )
  ;                             ( *4: game data* )
  ABUFF                         ( *3: addr of text GROB on stack* )
  #E ONE MAKEGROB
  DO>STR                        ( *2: vars string on stack* )
  ROT                           ( *1: bint* )

CODE
	CLRST                   Clear all flags (0-11)
	GOSBVL  =POP#           (#06641h) get the bint
	?A=0    A               Old game?
	GOYES   START           Yes, save RPL pointers
	ST=1    5               Else, this allows first ball to be deleted

START   GOSBVL  =SAVPTR         (#0679Bh) save RPL pointers

	A=DAT1  A               data string pointer
	A=A+CON A,10            skip past prologue, length
	R0=A.F  A               store in R0

	D1=D1+  5               Level 2
	A=DAT1  A               address of ABUFF
	LC(5)   20              skip past prologue, length, dims
	C=C+A   A               C[A] = address of text GROB
	GOSUBL  STO_00          save in storage string

	D1=D1+  5               Level 3
	C=DAT1  A               address of game data
	C=C+CON A,10            skip past prologue, length
	GOSUBL  STO_03          save data address

* X coords

	GOSUB   XSTO            get address of X coords
	NIBHEX	22000
	NIBHEX	B2000
	NIBHEX	43000
	NIBHEX	D3000
	NIBHEX	64000
	NIBHEX	F4000
	NIBHEX	85000
XSTO    C=RSTK
	GOSUBL  STO_04          store X address

* Y coords

	GOSUB   YSTO            get address of Y coords
	NIBHEX	10000
	NIBHEX	A0000
	NIBHEX	31000
	NIBHEX	C1000
	NIBHEX	52000
	NIBHEX	E2000
	NIBHEX	73000
YSTO    C=RSTK
	GOSUBL  STO_05          store Y address

****************************************************************************
* Draw the initial display.
****************************************************************************

* Pause for a short while.

	LCHEX   1FFFF
W1      C=C-1   A
	GONC    W1

* Get current contrast value and save it for screen restoration later.

	D0=(5)  #00101          Address for contrast
	A=DAT0  B               Get current contrast
	R4=A.F  B               Save for restoration

* Fade out screen.

CD1     LCHEX   01FFF
CD2     C=C-1   A
	GONC    CD2
	?A=0    B               Contrast down?
	GOYES   SETUP           Yes, go draw the board
	A=A-1   B               Else, decrement contrast
	DAT0=A  B
	GOTO    CD1

* Draw the board.

SETUP   GOSUBL  CLS             First clear the screen
	C=0     S
	D=C     S               Initialise row counter
	LA(5)   6               Max row/col #
	ASRC
	R1=A.F  S               Save max # for later
NXTROW  C=0     S
	B=C     S               Initialise col counter
NXTCOL  A=0     W
	A=B	S
	ASLC                    A[A] = col#
	R1=A.F  A
	C=0	W
	C=D     S
	CSLC                    C[A] = row#
	R2=C.F  A
	GOSUBL  BOARD           What does the location contain?
	?C=0    P               Is it a null location?
	GOYES   NXTPOS          Yes, move to next position
	C=R1.F  A               C[A] = col#
	GOSUBL  STO_01          save it
	C=R2.F  A               C[A] = row#
	GOSUBL  STO_02          save it
	A=R1.F  A
	C=R2.F  A
	GOSUBL  BOARD           What does the location contain?
	LA(1)   1
	?C=A    P               Is it a space?
	GOYES   SPC             Yes, display it
	GOSUBL  BALL            Else display a ball
	GOTO    NXTPOS
SPC     GOSUBL  SPACE
NXTPOS  B=B+1   S               Next column
	C=R1.F	S
	C=C-B	S
	GONC    NXTCOL          Loop back if not finished
	D=D+1   S               Else, do next row
	C=R1.F	S
	C=C-D	S
	GOC     CONUP           Restore the contrast
	GOLONG  NXTROW          Else, loop back

* Bring up the screen.

CONUP   D0=(5)  #00101          address for contrast
	A=R4.F  B               Recall contrast value
	B=A     B
	A=0     B               Contrast currently at 0
CU1     LCHEX   01FFF
CU2     C=C-1   A
	GONC    CU2
	?A=B    B               Contrast reset?
	GOYES   PLAY            Yes, go play game
	A=A+1   B               Else, increment contrast
	DAT0=A  B
	GOTO    CU1

****************************************************************************
* Play starts here.
****************************************************************************

PLAY    LC(5)   3
	A=C     A
	GOSUBL  STO_01          | Always start in the centre
	GOSUBL  STO_02          |
	C=A     A
	GOSUBL  BOARD
	LA(1)   1
	?C=A    P               Space?
	GOYES   P1              Yes
	ST=0    2               No, ball
	GOTO    P2
P1      ST=1    2               Space
P2      ST=1    1               Set display flag
	ST=0    3               Cursor movement

****************************************************************************
* Key handler - Uses low power keyboard scans.
* ( See Jim Donnelly's book )
****************************************************************************

KEYS    GOSUB   WaitKeyUp
	GOSBVL  =DisableIntr    (#26791h)
	GOSUB   BusyOff

LiteSlp D1=(5)  =TIMERCTRL.1    (#0012Eh) Set timer 1 to wake up CPU
	LC(1)	4
	DAT1=C	P
	D1=(2)  =TIMER1         (#00137h) Set a 3/16 second delay
	LC(1)   3
	DAT1=C	P
	LCHEX   1FF             Preload keyboard row lines
	OUT=C
Wait    SHUTDN                  WAIT FOR INTERRUPTS
	LCHEX   1FF             Load row lines
	OUT=C
	GOSBVL  =CINRTN         (#00212h) Read column lines
	LAHEX   0803F           Mask for all column lines
	A=A&C	A
	?A#0    A               Was a key pressed?
	GOYES   GetKey          Yes, check which ones
	D1=(2)  =TIMERCTRL.1    No, check timer control
	C=DAT1  X               Read timer status
	?CBIT=0 3               Has it expired?
	GOYES   Wait            No, go back to sleep
	GOSUB   Blink           Blink the active sprite
	GOTO    LiteSlp         Restart timer and sleep
*
GetKey  GOSBVL  =Debounce       (#2678Ah)
	?A#0	W		Was a key pressed?
	GOYES	GotKey
	GOTO    LiteSlp         Reset timer and sleep
*
GotKey  GOSUB   KeyNum
	GOSUB   WaitKeyUp       Wait for key to be released

****************************************************************************
* Find which key was pressed.
*
ON?     LC(2)   1               Key# for ON
	?B#C    B
	GOYES   ENTER?
	GOTO    DoON
*
ENTER?  LC(2)   57              Key# for ENTER
	?B#C    B
	GOYES   SPACE?
	GOTO    DoENTER
*
SPACE?  LC(2)   49              Key# for SPACE?
	?B#C    B
	GOYES   LEFT?
	GOTO    DoSPACE
*
LEFT?   LC(2)   11              Key# for LEFT
	?B#C    B
	GOYES   RIGHT?
	ST=0    6
	GOTO    DoLeftRight
*
RIGHT?  LC(2)   9               Key# for RIGHT
	?B#C    B
	GOYES   UP?
	ST=1    6
	GOTO    DoLeftRight
*
UP?     LC(2)   12              Key# for UP
	?B#C    B
	GOYES   DOWN?
	ST=0    6
	GOTO    DoUpDown
*
DOWN?   LC(2)   10              Key# for DOWN
	?B#C    B
	GOYES   TWO?
	ST=1    6
	GOTO    DoUpDown
*
TWO?    LC(2)   42              Key# for 2
	?B#C    B
	GOYES   FOUR?
	ST=1    6
	GOTO    DoUpDown
*
FOUR?   LC(2)   35              Key# for 4
	?B#C    B
	GOYES   SIX?
	ST=0    6
	GOTO    DoLeftRight
*
SIX?    LC(2)   51              Key# for 6
	?B#C    B
	GOYES   EIGHT?
	ST=1    6
	GOTO    DoLeftRight
*
EIGHT?  LC(2)   44              Key# for 8
	?B#C    B
	GOYES   FIVE?
	ST=0    6
	GOTO    DoUpDown
*
FIVE?   LC(2)   43              Key# for 5
	?B#C    B
	GOYES   LSleep
	GOTO    DoFIVE
*
LSleep  GOTO    LiteSlp

****************************************************************************
* Key Actions.
*
DoON    GOSUB   WaitKeyUp
	GOSBVL  =AllowIntr      (#26767h)
	GOSUB	BusyOn
	GOVLNG  =GPPushFLoop    (#3524Fh) restore RPL, push FALSE on stack
*
DoENTER ?ST=1   5               If storing a game that hasn't been played
	GOYES   DoON            abort it
	GOSUB   WaitKeyUp
	GOSBVL  =AllowIntr      (#26767h)
	GOSUB	BusyOn
	GOVLNG  =GPPushTLoop    (#35236h) restore RPL, push TRUE on stack
*
DoSPACE ?ST=1   5
	GOYES   DL1
	GOTO    LiteSlp
DL1     ?ST=1   1
	GOYES   DL2
	GOSUB   Blink
DL2     ST=1    3
	GOSUB   REMOVE
	ST=0    5
	ST=0    3
	GOTO    LiteSlp
*
DoFIVE  ?ST=0   3               Cursor mode?
	GOYES   E1              Yes, go set move mode
	?ST=1   1
	GOYES   E2
	GOSUB   Blink
E2      ST=0    3               Cursor mode on
	GOTO    LiteSlp
E1      ?ST=0   2               Ball?
	GOYES   E4              Yes, flash it
	GOTO    LiteSlp         Else, stay in cursor mode
E4      ?ST=1   1
	GOYES   E3
	GOSUB   Blink
E3      ST=1    3               Move mode on
	GOTO    LiteSlp
*
DoLeftRight
	?ST=1   1               Is sprite displayed?
	GOYES   LR1             Yes, do nothing
	GOSUB   TurnOn          Else, display it
LR1     GOSUB   RCL_01          Get current col#
	?ST=1   6               Moving right?
	GOYES   LR10            Yes, go and move
	A=0     A               A[A] = min col#
	GOTO    LR11
LR10    LA(5)   6               A[A] = max col#
LR11    ?C#A    A               Are you at the boundary?
	GOYES   LR2             No, continue
	ST=0    3
	GOTO    LiteSlp         Else, exit
LR2     ?ST=1   6
	GOYES   LR12
	C=C-1   A               Decrement col# to look ahead
	GOTO    LR13
LR12    C=C+1   A               Increment col# to look ahead
LR13    R1=C.F  A               Save it
	A=C     A               A[A] = col#
	GOSUB   RCL_02          C[A] = row#
	GOSUB   BOARD           See what's there
	?C#0    P               Is it a no-go region?
	GOYES   LR3             No, continue
	ST=0    3
	GOTO    LiteSlp         Else, exit
LR3     LA(1)   1
	?C=A    P               Is it a space?
	GOYES   LR4
	ST=0    2               Ball
	?ST=0   3
	GOYES   LR5
	GOTO    LRMOVE          Make your move
LR5     C=R1.F  A               Recall new col#
	GOSUB   STO_01          Update it
	ST=0    3
	GOTO    LiteSlp
LR4     ST=1    2               Space
	C=R1.F  A               Recall new col#
	GOSUB   STO_01          Update it
	ST=0    3
	GOTO    LiteSlp
*
LRMOVE  GOSUB   RCL_02          C[A] = row#
	LA(5)   2
	?C<A    A               row# < 2?
	GOYES   LR6
	LA(5)   4
	?C>A    A               row# > 4?
	GOYES   LR6
	GOSUB   RCL_01
	?ST=1   6
	GOYES   LR14
	LA(5)   1
	?C>A    A               if r>=2 and r<=4, then must have c>1
	GOYES   LR7
	GOTO    LR5
LR14    LA(5)   5
	?C<A    A               if r>=2 and r<=4, then must have c<5
	GOYES   LR7
	GOTO    LR5
LR6     GOSUB   RCL_01          C[A] = col#
	?ST=1   6
	GOYES   LR15
	LA(5)   4
	?C=A    A               col# = 4?
	GOYES   LR7
	GOTO    LR5
LR15    LA(5)   2
	?C=A    A
	GOYES   LR7
	GOTO    LR5

* What are you trying to move?

LR7     GOSUB   RCL_01          Get current col#
	A=C     A               A[A] = col#
	GOSUB   RCL_02          C[A] = row#
	GOSUB   BOARD           Access the board
	LA(1)   2
	?C=A    P               Is there a ball?
	GOYES   LR8             Yes, find out where you are moving it
	GOTO    LR5             No, just move to the next col

* Where are you trying to move it?

LR8     GOSUB   RCL_01
	?ST=1   6
	GOYES   LR16
	C=C-CON A,2
	GOTO    LR17
LR16    C=C+CON A,2
LR17    A=C     A               A[A] = col#
	GOSUB   RCL_02          C[A] = row#
	GOSUB   BOARD           Access the board
	LA(1)   1
	?C=A    P               Is it a space?
	GOYES   LR9             Yes, do the move
	GOTO    LR5             No, just move to the next col

* Make your move.

LR9     GOSUB   REMOVE          Remove ball from current location
	GOSUB   RCL_01
	?ST=1   6
	GOYES   LR18
	C=C-1   A
	GOTO    LR19
LR18    C=C+1   A
LR19    RSTK=C
	GOSUB   STO_01          Move to next col
	GOSUB   REMOVE          Remove ball
	C=RSTK
	?ST=1   6
	GOYES   LR20
	C=C-1   A
	GOTO    LR21
LR20    C=C+1   A
LR21    GOSUB   STO_01          Move to dest col
	GOSUB   PLACE           Place ball here
	ST=0    3
	GOTO    LiteSlp         That's it
*
DoUpDown
	?ST=1   1               Is ball displayed?
	GOYES   UD1             Yes, do nothing
	GOSUB   TurnOn          Else, display it
UD1     GOSUB   RCL_02          Get current row#
	?ST=1   6
	GOYES   UD10
	A=0     A
	GOTO    UD11
UD10    LA(5)   6
UD11    ?C#A    A               Are you at the boundary?
	GOYES   UD2             No, continue
	ST=0    3
	GOTO    LiteSlp         Else, exit
UD2     ?ST=1   6
	GOYES   UD12
	C=C-1   A               Decrement row#
	GOTO    UD13
UD12    C=C+1   A               Increment row#
UD13    R1=C.F  A               Save it
	A=C     A               A[A] = row#
	GOSUB   RCL_01          C[A] = col#
	ACEX    A               Swap
	GOSUB   BOARD           See what's there
	?C#0    P               Is it a no-go region?
	GOYES   UD3             No, continue
	ST=0    3
	GOTO    LiteSlp         Else, exit
UD3     LA(1)   1
	?C=A    P               Is it a space?
	GOYES   UD4
	ST=0    2               Ball
	?ST=0   3
	GOYES   UD5
	GOTO    UDMOVE
UD5     C=R1.F  A               Recall new row#
	GOSUB   STO_02          Update it
	ST=0    3
	GOTO    LiteSlp
UD4     ST=1    2               Space
	C=R1.F  A               Recall new row#
	GOSUB   STO_02          Update it
	ST=0    3
	GOTO    LiteSlp
*
UDMOVE  GOSUB   RCL_01          C[A] = col#
	LA(5)   2
	?C<A    A               col# < 2?
	GOYES   UD6
	LA(5)   4
	?C>A    A               col# > 4?
	GOYES   UD6
	GOSUB   RCL_02
	?ST=1   6
	GOYES   UD14
	LA(5)   1
	?C>A    A               if c>=2 and c<=4, then must have r>1
	GOYES   UD7
	GOTO    UD5
UD14    LA(5)   5
	?C<A    A               if c>=2 and c<=4, then must have r<5
	GOYES   UD7
	GOTO    UD5
UD6     GOSUB   RCL_02          C[A] = row#
	?ST=1   6
	GOYES   UD15
	LA(5)   4
	?C=A    A               row# = 4?
	GOYES   UD7
	GOTO    UD5
UD15    LA(5)   2
	?C=A    A               row# = 2?
	GOYES   UD7
	GOTO    UD5

* What are you trying to move?

UD7     GOSUB   RCL_01          Get current col#
	A=C     A               A[A] = col#
	GOSUB   RCL_02          C[A] = row#
	GOSUB   BOARD           Access the board
	LA(1)   2
	?C=A    P               Is there a ball?
	GOYES   UD8             Yes, find out where you are moving it
	GOTO    UD5             No, just move to the next row

* Where are you trying to move it?

UD8     GOSUB   RCL_01
	A=C     A               A[A] = col#
	GOSUB   RCL_02          C[A] = row#
	?ST=1   6
	GOYES   UD16
	C=C-CON A,2
	GOTO    UD17
UD16    C=C+CON A,2
UD17    GOSUB   BOARD           Access the board
	LA(1)   1
	?C=A    P               Is it a space?
	GOYES   UD9             Yes, do the move
	GOTO    UD5             No, just move to the next row
*
UD9     GOSUB   REMOVE          Remove ball from current location
	GOSUB   RCL_02
	?ST=1   6
	GOYES   UD18
	C=C-1   A
	GOTO    UD19
UD18    C=C+1   A
UD19    RSTK=C
	GOSUB   STO_02          Move to next row
	GOSUB   REMOVE          Remove ball
	C=RSTK
	?ST=1   6
	GOYES   UD20
	C=C-1   A
	GOTO    UD21
UD20    C=C+1   A
UD21    GOSUB   STO_02          Move to dest row
	GOSUB   PLACE           Place ball here
	ST=0    3
	GOTO    LiteSlp         That's it

****************************************************************************
**                           SUBROUTINES
****************************************************************************

****************************************************************************
* Wait for keys to be released.
****************************************************************************

WaitKeyUp
	LCHEX   1FF             Set row lines
	OUT=C
	GOSBVL  =CINRTN         Read column lines
	LAHEX   0803F           Mask for column lines
	A=A&C	A
	?A#0    A               Were any keys down?
	GOYES   WaitKeyUp       Yes, go scan again
	RTN                     No, return

****************************************************************************
*  Calculate the key number.
****************************************************************************

KeyNum	B=0	A
NextNib	?A#0	P
	GOYES	TestBits
	B=B+CON	B,4
	ASR	W
	GONC    NextNib
*
TestBits
	B=B+1	B
	SB=0
	ASRB.F	P
	?SB=0
	GOYES	TestBits
	RTN

****************************************************************************
* Blink the active sprite. Subroutine SPRITE performs the RTN.
****************************************************************************

Blink   ?ST=0   1               Sprite not displayed?
	GOYES   TurnOn          Display it
	ST=0    1               Else, turn off status and remove it
	GONC    BL1
TurnOn
	ST=1    1               Set status
BL1     ?ST=0   3               Cursor mode
	GOYES   BL3             Yes, go display cursor
	?ST=1   2               Else, check if ball or space
	GOYES   BL2
	GOTO    BALL            Display/remove ball
BL2     GOTO    SPACE           Display/remove space
BL3     GOTO    CURSOR          Display/remove cursor

****************************************************************************
* Routine to remove a ball.
****************************************************************************

REMOVE  ?ST=0   1               Sprite not displayed?
	GOYES   RM1             Yes, continue
	GOSUB   BALL            Else, remove it
RM1     GOSUB   RCL_01          Col#
	A=C     A
	GOSUB   RCL_02          Row#
	GOSUB   BOARD           Access the board
	LC(1)   1
	DAT0=C  P               Write 'space' to the board
	ST=1    2               Adjust display flag
	GOSUB   TurnOn          Display the sprite
	RTN

****************************************************************************
* Routine to place a ball.
****************************************************************************

PLACE   ?ST=0   1               Sprite not displayed?
	GOYES   PL1             Yes, continue
	GOSUB   SPACE           Else, remove it
PL1     GOSUB   RCL_01          Col#
	A=C     A
	GOSUB   RCL_02          Row#
	GOSUB   BOARD           Access the board
	LC(1)   2
	DAT0=C  P               Write 'space' to the board
	ST=0    2               Adjust display flag
	GOSUB   TurnOn          Display the sprite
	RTN

****************************************************************************
* Clear the screen.
****************************************************************************

CLS     GOSUBL  RCL_00
	D0=C                    D0 -> screen
	LCHEX   87              C[B] = (880h/10h)-1
	A=0     W
CL1     DAT0=A  W               Write 16 zeros to the screen each time
	D0=D0+  16
	C=C-1   B
	GONC    CL1
	RTN

****************************************************************************
* Turn off the busy annunciator.
****************************************************************************

BusyOff D0=(5)  (=ANNCTRL)+1    (#0010Bh+1) Point at annunciator nibble
	C=DAT0  P               Read nibble
	CBIT=0  0               Clear annunciator bit
	DAT0=C  P               Write nibble back
	RTN

****************************************************************************
*  Turn on the busy annunciator.
****************************************************************************

BusyOn  D0=(5)  (=ANNCTRL)+1    Point at annunciator nibble
	C=DAT0  P               Read nibble
	CBIT=1  0               Clear annunciator bit
	DAT0=C  P               Write nibble back
	RTN

****************************************************************************
*
* STORE variable
*
* Each variable is numbered, and 5 nibbles in length.
* STO_00 will store C[A] into variable #0.
* All registers are preserved except for C[A] and A[W],
* but A[A] is preserved.
*
****************************************************************************

STO_00  RSTK=C
	LC(5)    0
	GOTO     SC
STO_01  RSTK=C
	LC(5)    1
	GOTO     SC
STO_02  RSTK=C
	LC(5)    2
	GOTO     SC
STO_03  RSTK=C
	LC(5)    3
	GOTO     SC
STO_04  RSTK=C
	LC(5)    4
	GOTO     SC
STO_05  RSTK=C
	LC(5)    5
	GOTO     SC
STO_06  RSTK=C
	LC(5)    6
	GOTO     SC
STO_07  RSTK=C
	LC(5)    7
	GOTO     SC
STO_08  RSTK=C
	LC(5)    8
	GOTO     SC
STO_09  RSTK=C
	LC(5)    9
	GOTO     SC
STO_10  RSTK=C
	LC(5)    10
	GOTO     SC
STO_11  RSTK=C
	LC(5)    11
	GOTO     SC
STO_12  RSTK=C
	LC(5)    12
	GOTO     SC
STO_13  RSTK=C
	LC(5)    13

****************************************************************************
*
* When SC is called, C[A] must contain the variable number that you
* would like to store, and RSTK should contain the value that will be
* put into it.  All regs are preserved except for A[W].  A[W] is used,
* but A[A] is preserved.  One level of RSTK are used.  (Actually,
* SC doesn't use an RSTK level, but the calling routine above does.
*
****************************************************************************

SC      GOSBVL  =ASLW5          (#26E71h) Shift A[W] left 5 nibbles
*                               (preserve A[A])
	AD0EX                   Get D0 into A[A]
	GOSBVL  =ASLW5          Shift A[W], preserve D0.
	A=C     A               Get var #
	C=C+C   A               | Multiply C[A] by 5.
	C=C+C   A               | This is because each var is 5 nibbles
	C=C+A   A               | in length.  This gets the correct offset.
	A=R0.F  A               Get address of data string.
	A=A+C   A               Add offset to the var #.
	D0=A                    Put into D0.
	C=RSTK                  Get value off RSTK
	DAT0=C  A               Store value into Variable area.
	GOSBVL  =ASRW5          (#26E60h) Put orig D0 into A[A]
	D0=A                    Restore D0
	GOSBVL  =ASRW5          Restore A[A].
	RTN

****************************************************************************
*
* RECALL variable
*
* Each variable is numbered, and 5 nibbles in length.
* RCL_00 will recall variable #0 in C[A].
* All registers are preserved except for C[A] and A[W],
* but A[A] is preserved.
*
****************************************************************************

RCL_00  LC(5)    0
	GOTO     RC
RCL_01  LC(5)    1
	GOTO     RC
RCL_02  LC(5)    2
	GOTO     RC
RCL_03  LC(5)    3
	GOTO     RC
RCL_04  LC(5)    4
	GOTO     RC
RCL_05  LC(5)    5
	GOTO     RC
RCL_06  LC(5)    6
	GOTO     RC
RCL_07  LC(5)    7
	GOTO     RC
RCL_08  LC(5)    8
	GOTO     RC
RCL_09  LC(5)    9
	GOTO     RC
RCL_10  LC(5)    10
	GOTO     RC
RCL_11  LC(5)    11
	GOTO     RC
RCL_12  LC(5)    12
	GOTO     RC
RCL_13  LC(5)    13

****************************************************************************
*
* When RC is called, C[A] must contain the variable number that you
* would like to retrieve.  All regs are preserved even though A[A] and
* D0 are used.  One level of RSTK is used.
*
****************************************************************************

RC      ACEX    A               Swap A and C
	CD0EX                   Swap D0 and C
	RSTK=C                  Put the orig D0 value onto RSTK
	C=A     A               Copy var # into C.
	C=C+C   A               | Multiply C[A] by 5.
	C=C+C   A               | This is because each var is 5 nibbles
	C=C+A   A               | in length.  This gets the correct offset.
	A=R0.F  A               Get address of data string.
	A=A+C   A               Add the offset to the var #.
	AD0EX                   Restore A[A] (it was in D0).
*                               At the same time, this puts the correct
*                               Var # RAM address into D0 for use.
	C=DAT0  A               Read the 5 nibble var into C[A].
	CD0EX                   Put the value into D0
	C=RSTK                  Get orig D0 value from RSTK
	CD0EX                   Restore orig D0 value, get var into C[A].
	RTN

****************************************************************************
* This routine checks the state of a given location. Enter with the col#
* in A[A] and the row# in C[A]. The result, returned in C[P], is as
* follows:-
*    0 - Out of bounds
*    1 - Location contains a space
*    2 - Location contains a ball
****************************************************************************

BOARD   ACEX    A
	RSTK=C                  STO col#
	C=A     A               |
	C=C+C   A               |
	C=C+C   A               | row#*7
	C=C+C   A               |
	C=C-A   A               |
	A=C	A
	C=RSTK                  RCL col#
	A=A+C   A               Form offset (7 * row# + col#)
	GOSUB   RCL_03          Get game data addr
	C=C+A   A               Add offset
	D0=C                    D0 points to the required element
	C=DAT0  P               Get element
	RTN

****************************************************************************
* Get screen position of sprite.
****************************************************************************

GETX    R1=A.F  A               Preserve A[A]
	GOSUB   RCL_01          Get row#
	A=C     A
	GOSUB   RCL_04          Get ptr to X coords table
	GOTO    XY

GETY    R1=A.F  A               Preserve A[A]
	GOSUB   RCL_02          Get col#
	A=C     A
	GOSUB   RCL_05

XY      D0=C
	C=A     A               Copy element # into C.
	C=C+C   A               | Multiply C[A] by 5.
	C=C+C   A               | This is because each var is 5 nibbles
	C=C+A   A               | in length.  This gets the correct offset.
	AD0EX                   Get address of table.
	A=A+C   A               Add the offset to the address.
	AD0EX                   Restore A[A].
	C=DAT0  A               Read the 5 nibble var into C[A].
	A=R1.F  A               Restore A[A]
	RTN

****************************************************************************
* Display cursor. There's room for improvement here. The sprite drawing
* subroutine currently copes with a max width of 1 byte.
****************************************************************************

CURSOR  GOSUB   CU              push sprite address
	NIBHEX  3030000000
	NIBHEX  0000003030
CU      C=RSTK                  pop sprite address
	R2=C.F  A

	D1=C                    D1 = sprite pointer
	LC(1)   9               10 lines (0-9) to display
	D=C     P               D[P] = line counter
	GOSUB   GETX
	C=C-1   A               X position of cursor
	R3=C.F  A
	A=C     A               save in A[A]
	GOSUB   GETY
	C=C-1   A               Y position of cursor in C[A]
	R4=C.F  A
	RSTK=C
	GOSUB   RCL_00          get screen address
	D0=C                    D0 = screen pointer
	C=RSTK
	GOSUB   SPRITE

	C=R2.F  A
	D1=C                    D1 = sprite pointer
	LC(1)   9               10 lines (0-9) to display
	D=C     P               D[P] = line counter
	C=R3.F  A
	LA(5)   8
	A=A+C   A               X position of cursor
	GOSUB   RCL_00          get screen address
	D0=C                    D0 = screen pointer
	C=R4.F  A               Recall Y pos
	GOTO    SPRITE

****************************************************************************
* Display ball or cup.
****************************************************************************

BALL    GOSUB   BL              push sprite address
	NIBHEX  C3A79FDF        bitmap of ball (8x8)
	NIBHEX  FFFFE7C3
BL      C=RSTK                  pop sprite address
	GOTO    BS1

SPACE   GOSUB   SP              push sprite address
	NIBHEX  C3241818        bitmap of space (8x8)
	NIBHEX  181824C3
SP      C=RSTK                  pop sprite address

BS1     D1=C                    D1 = sprite pointer
	LC(1)   7               8 lines (0-7) to display
	D=C     P               D[P] = line counter
	GOSUB   GETX            get X position of ball
	A=C     A               save in A[A]
	GOSUB   GETY            get Y position of ball in C[A]
	RSTK=C
	GOSUB   RCL_00          get screen address
	D0=C                    D0 = screen pointer
	C=RSTK

****************************************************************************
* Display sprite. It XOR's the sprite onto the screen.
*
* D0   = screen pointer
* D1   = sprite pointer
* A[A] = X position
* C[A] = Y position
* D[P] = Height of sprite
****************************************************************************

SPRITE  C=C+C   A
	B=C	A
	CSL	A
	B=B+C   A               34*Y
	CD0EX
	C=C+B	A
	B=A	A
	ASRB.F	A
	ASRB.F  A               X/4
	C=C+A   A
	D0=C                    D0 = @screen + X/4 + 34*Y
NXTLIN  A=B     A               A[A]=X
	C=0     X
	C=DAT1  B               Get line of sprite
	?ABIT#1 0               Is X MOD 4 = 1 or 3 ?
	GOYES   SP1             Jump if not
	C=C+C   X               Else shift sprite 1 bit left
SP1     ?ABIT#1 1               Is X MOD 4 = 2 or 3  ?
	GOYES   SP2             Jump if not
	C=C+C   X               | Else shift sprite 2 bits left
	C=C+C   X               |
SP2     A=DAT0  X
	R1=A.F  X               |
	A=A!C   X               |
	AR1EX.F X               |
	C=C&A   X               | A = A XOR C
	C=-C-1  X               |
	A=R1.F  X               |
	A=A&C   X               |
	DAT0=A  X
	D1=D1+  2               Next sprite line
	D0=D0+  16              |
	D0=D0+  16              | Next screen line
	D0=D0+  2               |
	D=D-1   P
	GONC    NXTLIN
	RTN
ENDCODE

  TURNMENUON                    ( *restore menus* )
  RECLAIMDISP			( *restore stack display* )

* The following actions depend on whether ON or ENTER was pressed. If ENTER,
* the program data is saved before exit, else, if ON, the data is dropped
* and any previously saved data purged.

  ITE
  ::                            ( *ENTER was pressed* )
    2DROP
    ' ID SOLpar XEQSTOID
  ;
  ::                            ( *ON was pressed* )
    3DROP
    ' ID SOLpar ?PURGE_HERE
  ;
;

****************************************************************************
****************************************************************************
